REBOL [
	Name: 		"NT-Services Scheme"
	Purpose:	"NT-Services support library"
	Author:		"Nenad Rakocevic / SOFTINNOV"
	Email:		nr@softinnov.com
	File: 		%scm-scheme.r
	Version:	1.0.3
	Date:		05/08/2009
	Note: 		"For Windows 2000/XP/Vista only"
	Needs: 		[library]
	
	History:	{
	
		1.0.3 - 05/08/2009
			o Install function now checks for presence of 'title property. (Thanks to Titus Barik)
	
		1.0.2 - 09/03/2009
			o Fixed a typo in 'running? function preventing it from returning the correct result.
			  (Thanks to David for finding and reporting it).
	
		1.0.1 - 23/02/2008 
			o In nt-service API, 'run function renamed to 'attach
			o Function 'write can now be used on scm:// ports.
	}
]

;-- inlined version of our %os-tools.r library
if not value? 'runtime-libs [
	runtime-libs: []

	DWORD: make struct! d-word: [
		low [integer!]
		high [integer!]
	] none
	
	make-lpDWORD: does [make struct! DWORD [0 0]]
	
	zero-buffer: func [len [integer!]][
		head insert/dup make string! len null len
	]
	
	address-of?: func [str [string!]][
		third make struct! [p [string!]] reduce [str]
	]

	load-thru-library: func [[catch] lib [file!]][
		any [
			select runtime-libs lib
			all [
				repend runtime-libs [
					lib
					lib: load/library lib
				]
				lib
			]
		]
	]
]

make root-protocol [
	scheme: 'SCM

	sw: system/words
	
	open*: 	 	get in sw 'open
	copy*: 	 	get in sw 'copy
	insert*: 	get in sw 'insert
	find*:	 	get in sw 'find
	remove*: 	get in sw 'remove
	pick*: 	 	get in sw 'pick
	close*:  	get in sw 'close
	get-modes*: get in sw 'get-modes
	query*:		get in sw 'query
	
	net-log: 	get in net-utils 'net-log

	open-proto: none
	port-flags: system/standard/port-flags/pass-thru
	
	defs: [
		states [
			stopped				1
			start-pending		2
			stop-pending		3
			running				4
			continue-pending	5
			pause-pending		6
			paused				7		
		]
		start [
			boot				0
			system				1
			auto				2
			manual				3
			disabled			4
		]
		errors [
			ignore				0
			normal				1
			severe				2
			critical			3
		]
		control [
			start				-
			stop				1
			pause 				2
			continue			3
		]
		types [
			kernel-driver		1
			file-system-driver	2
			win32-own-process	16
			win32-share-process	32
		]
		filters [
			active				1
			inactive			2
			all-states			3
			driver				11
			win32				48
		]
		misc [
			interactive			256
			manager-all-access	983103
			svc-all-access		983551
			delete-right		65536
			no-change			-1
		]
		config [
			desc				1
		]
	]
	
	kernel32: load-thru-library %kernel32.dll
	advapi32: load-thru-library %advapi32.dll

	; === General API ===
	
	int: :to-integer
	
	FORMAT_MESSAGE_FROM_SYSTEM:	   	int #{00001000}
	FORMAT_MESSAGE_IGNORE_INSERTS:	int #{00000200}
	ERROR_ACCESS_DENIED: 			5
	CSIDL_DESKTOPDIRECTORY:			int #{00000010}
	CSIDL_COMMON_APPDATA:			int #{00000023}
	SHGFP_TYPE_CURRENT:				0
	fmt-msg-flags: 
		FORMAT_MESSAGE_FROM_SYSTEM
		or FORMAT_MESSAGE_IGNORE_INSERTS
	
	CloseHandle: make routine! [
		hObject	[integer!]
		return: [integer!]
	] kernel32 "CloseHandle"
	
	GetLastError: make routine! [
		return: [integer!]
	] kernel32 "GetLastError"
	
	FormatMessage: make routine! [
		dwFlags		 [integer!]
		lpSource	 [integer!]
		dwMessageId  [integer!]
		dwLanguageId [integer!]
		lpBuffer	 [string!]
		nSize		 [integer!]
		Arguments	 [integer!]
		return:		 [integer!]
	] kernel32 "FormatMessageA"

	; === Service API ===

	ERROR_SERVICE_DOES_NOT_EXIST: 1060
	
	SERVICE_STATUS: make struct! struct-service-status: [
		dwServiceType			  [integer!]
		dwCurrentState			  [integer!]
		dwControlsAccepted 		  [integer!]
		dwWin32ExitCode			  [integer!]
		dwServiceSpecificExitCode [integer!]
		dwCheckPoint			  [integer!]
		dwWaitHint				  [integer!]
	] none
	
	enum-service-status: [
		[save]
		lpServiceName	[string!]
		lpDisplayName	[string!]
	]
	append enum-service-status struct-service-status
	
	QUERY_SERVICE_CONFIG: make struct! struct-query-config: [
		[save]
		dwServiceType			[integer!]
		dwStartType				[integer!]
		dwErrorControl			[integer!]
		lpBinaryPathName		[string!]
		lpLoadOrderGroup		[string!]
		dwTagId					[integer!]
		lpDependencies			[string!]
		lpServiceStartName		[string!]
		lpDisplayName			[string!]
	] none
	
	OpenSCManager: make routine! [
		lpMachineName	[integer!]
		lpDatabaseName	[integer!]
		dwDesiredAccess	[integer!]
		return:   		[integer!]
	] advapi32 "OpenSCManagerA"
	
	CreateService: make routine! [
		hSCManager   		[integer!]
		lpServiceName		[string!]
		lpDisplayName   	[string!]
		dwDesiredAccess		[integer!]
	  	dwServiceType		[integer!]
		dwStartType			[integer!]
	 	dwErrorControl		[integer!]
	  	lpBinaryPathName 	[string!]
		lpLoadOrderGroup 	[integer!]
		lpdwTagId			[integer!]
		lpDependencies		[integer!]
		lpServiceStartName	[integer!]
		lpPassword			[integer!]
		return:   			[integer!]
	] advapi32 "CreateServiceA"
	
	DeleteService: make routine! [
		hService	[integer!]
		return:		[integer!]
	] advapi32 "DeleteService"
	
	OpenService: make routine! [
		hSCManager 		[integer!]
		lpServiceName 	[string!]
		dwDesiredAccess [integer!]
		return:  		[integer!]
	] advapi32 "OpenServiceA"
	
	CloseServiceHandle: make routine! [
		hSCObject	[integer!]
		return:		[integer!]
	] advapi32 "CloseServiceHandle"
	
	ControlService: make routine! compose/deep [
		hService	[integer!]
		dwControl	[integer!]
		lpServiceStatus [struct! [(struct-service-status)]]
		return: 	[integer!]
	] advapi32 "ControlService"
	
	StartService: make routine! [
		hService 			[integer!]
		dwNumServiceArgs	[integer!]
		lpServiceArgVectors [integer!]
		return: 			[integer!]
	] advapi32 "StartServiceA"
	
	QueryServiceStatus: make routine! compose/deep [
		hService		[integer!]
		lpServiceStatus [struct! [(struct-service-status)]]
		return: 		[integer!]
	] advapi32 "QueryServiceStatus"
	
	EnumServicesStatus: make routine! compose/deep [
		hSCManager			[integer!]
	 	dwServiceType 		[integer!]
	  	dwServiceState 		[integer!]
		lpServices			[string!]
		cbBufSize			[integer!]
	  	pcbBytesNeeded		[struct! [(d-word)]]
		lpServicesReturned	[struct! [(d-word)]]
		lpResumeHandle		[struct! [(d-word)]]
		return: 			[integer!]
	] advapi32 "EnumServicesStatusA"

	QueryServiceConfig: make routine! compose/deep [
		hService		[integer!]
		lpServiceConfig [string!]
		cbBufSize		[integer!]
 		pcbBytesNeeded	[struct! [(d-word)]]
 		return: 		[integer!]
	] advapi32 "QueryServiceConfigA"
	
	QueryServiceConfig2: make routine! compose/deep [
		hService		[integer!]
		dwInfoLevel		[integer!]
		lpBuffer		[string!]
		cbBufSize		[integer!]
		pcbBytesNeeded	[struct! [(d-word)]]
		return: 		[integer!]
	] advapi32 "QueryServiceConfig2A"
	
	ChangeServiceConfig: make routine! [
		hService			[integer!]
		dwServiceType		[integer!]
		dwStartType			[integer!]
		dwErrorControl		[integer!]
		lpBinaryPathName 	[string!]
		lpLoadOrderGroup 	[integer!]
		lpdwTagId			[integer!]
		lpDependencies		[integer!]
		lpServiceStartName	[string!]
		lpPassword			[string!]
		lpDisplayName   	[string!]
		return:   			[integer!]
	] advapi32 "ChangeServiceConfigA"
	
	ChangeServiceConfig2: make routine! [
		hService			[integer!]
		dwInfoLevel			[integer!]
		lpInfo				[struct! [s [string!]]]
		return:   			[integer!]
	] advapi32 "ChangeServiceConfig2A"

	; === Helper functions ====
	
	last-error: none
	svc-open-rights: defs/misc/delete-right or defs/misc/svc-all-access
	
	types-mask: 0
	foreach [n v] defs/types [types-mask: types-mask or v]
	
	any-running: reduce [
		defs/states/running
		defs/states/start-pending
		defs/states/continue-pending
	]
	
	svc-class: context [
		name: title: comment: command: type: start-type: interactive:
		error-control: user: pass: none
	]
	
	status-class: context [
		state: type: accepted: exit-code: interactive: none
	]
	
	get-error-msg: has [out][
		out: zero-buffer 256
		FormatMessage fmt-msg-flags 0 last-error: GetLastError 0 out 256 0
		trim/tail out
	]
	
	try*: func [body [block!] /quiet /local res][
		if all [zero? res: do body not quiet][
			make error! reform [
				mold first body "failed :" get-error-msg
			]
		]
		res
	]
	
	decode-flags: func [s [struct!] /state /type /error /start][
		first back case [
			state [find* defs/states s/dwCurrentState]
			error [find* defs/errors s/dwErrorControl]
			type  [find* defs/types  s/dwServiceType and types-mask]
			start [find* defs/start  s/dwStartType]
		]
	]
	
	encode-flags: func [symbol [word!] kind [word!]][
		select defs/:kind symbol
	]
	
	set-config: func [svc spec /local cmd routine value][
		routine: third :ChangeServiceConfig
		cmd: [
			ChangeServiceConfig svc - - - - 0 0 0 - - -
		]
		foreach [pos word default patch code][
			3	type:	 		-1	 -	types
			4	start-type:		-1	 -	start
			5	error-control:	-1	 -	errors
			6	command:		 0	 10	  -
			10	user:			 0	 18	  -
			11	pass:			 0	 20	  -
			12	title:			 0	 22	  -
			3	interactive:	-1   -    -
		][
			all [
				value: select spec word
				code <> '-
				value: encode-flags value code
			]
			either all [word = 'interactive positive? cmd/:pos][
				if value [
					poke cmd pos cmd/:pos or defs/misc/interactive
				]
			][
				if string? value [value: join value null]
				poke cmd pos any [value default]
			]
			if integer? patch [
				change routine/:patch pick [integer! string!] none? value 
			]
		]	
		try* cmd
	]
	
	get-config: func [svc host spec /local cmd buffer sz needed slot ts out][
		buffer: ""			;-- fake value just for first call
		sz: 0
		needed:	make-lpDWORD
		
		cmd: [QueryServiceConfig svc buffer sz needed]
		do cmd
		buffer: zero-buffer sz: needed/low
		try* cmd
		slot: make struct! struct-query-config none
		change/part ts: third slot buffer length? ts
		buffer: none
		out: make svc-class []
		out/name: host
		out/title: copy* slot/lpDisplayName
		out/command: copy* slot/lpBinaryPathName
		out/type: decode-flags/type slot
		out/start-type: decode-flags/start slot
		out/interactive: not zero? slot/dwServiceType and defs/misc/interactive
		out/error-control: decode-flags/error slot
		out/user: copy* slot/lpServiceStartName
		out
	]
	
	set-comment: func [svc spec /local buffer][
		buffer: make struct! [[save] s [string!]] reduce [
			copy* select spec first [comment:]
		]
		try* [
			ChangeServiceConfig2 svc defs/config/desc buffer
		]
	]
	
	get-comment: func [svc /local cmd buffer sz needed slot ts][
		buffer: ""			;-- fake value just for first call
		sz: 0
		needed:	make-lpDWORD
		
		cmd: [QueryServiceConfig2 svc defs/config/desc buffer sz needed]
		do cmd
		buffer: zero-buffer sz: needed/low
		try* cmd
		
		slot: make struct! [[save] s [string!]] none
		change/part ts: third slot buffer length? ts
		buffer: none
		all [slot/s trim/tail copy* slot/s]
	]
	
	list-services: func [
		port 
		/local
			buffer sz needed returned resume enum slot slot* len out
			f1 f2 pl p flag
	][
		buffer:   ""			;-- fake value just for first call
		sz: 0
		needed:	  make-lpDWORD
		returned: make-lpDWORD
		resume:	  make-lpDWORD
		
		flag: defs/filters
		pl: port/locals
		f1: any [
			do select [win32 flag/win32 drivers flag/driver] pl/filter1
			flag/win32 or flag/driver 
		]
		f2: any [
			do select [active flag/active inactive flag/inactive] pl/filter2
			flag/all-states
		]
		enum: [
			EnumServicesStatus pl/scm f1 f2 buffer sz needed returned resume
		]
		do enum
		buffer: zero-buffer sz: needed/low
		try* enum
		
		slot: make struct! enum-service-status none 
		len: length? slot*: third slot
		out: make block! returned/low
		p: buffer
		loop returned/low [
			change/part slot* p at p len
			append/only out new-line/all reduce [
				decode-flags/state slot
				slot/lpServiceName
				slot/lpDisplayName	
			] false
			p: at p len + 1
		]
		buffer: none
		new-line/all out true
	]
		
	install: func [scm spec /local srv start err type user pass cmd mode][
		spec: make svc-class spec
		if not all [spec/name spec/command spec/title][
			make error! "missing name, command or title in service spec"
		]
		if not start: select defs/start any [spec/start-type 'auto][
			make error! "invalid 'start-type"
		]
		if not err: select defs/errors any [spec/error-control 'ignore][
			make error! "invalid 'error-control"
		]
		type: any [spec/type defs/types/win32-own-process]
		if spec/interactive [type: type or defs/misc/interactive]
		
		cmd: [
			CreateService
				scm
				spec/name
				spec/title
				defs/misc/svc-all-access
				type
				start
				err
				spec/command
				0 0 0 
				user
				pass
		]
		user: pass: 0
		mode: 'integer!
		if spec/user [
			user: spec/user 
			pass: any [spec/pass ""]
			mode: 'string!
		]
		change pick third :CreateService 24 mode
		change pick third :CreateService 26 mode
		
		srv: try* cmd
		if not zero? srv [
			if spec/comment [
				set-comment srv compose [comment: (spec/comment)]
			]
			try* [CloseServiceHandle srv]
		]
	]
		
	control: func [svc cmd /local ss][
		if not find* defs/control cmd: first cmd [
			make error! "unknown service control command"
		]
		to-logic either cmd = 'start [
			try* [StartService svc 0 0]
		][
			ss: make struct! SERVICE_STATUS none
			try* [ControlService svc defs/control/:cmd ss]
		]
	]
	
	check-action: func [action [word!] /svc][
		if find* pick [
			[insert find]
			[set-modes get-modes remove]
		] to-logic svc action [
			make error! reform [
				"Port doesn't support"
				uppercase mold action
				"action"
			]
		]
	]
	
	;------- External API -------

	init: func [port spec][		
		if url? spec [net-utils/url-parser/parse-url port spec]
		port/url: spec
		port/locals: context [
			scm: none		; SCM handle
			svc: none		; service handle
			filter1: 'win32	; can be 'win32 | 'drivers | none
			filter2: none	; can be 'active | 'inactive | none
		]
		port
	]

	open: func [port /local svc pl][
		port/state/flags: port/state/flags or port-flags
		pl: port/locals
		
		pl/scm: try* [OpenSCManager 0 0 defs/misc/manager-all-access]
		if zero? pl/scm [make error! last-error]
		
		if port/host [
			svc: try* [OpenService pl/scm port/host svc-open-rights]
			if zero? svc [
				close port
				make error! last-error
			]
			pl/svc: svc
		]
		port/state/tail: 1 ;-- hack to enable 'remove calls
		port
	]
	
	insert: func  [port spec [block!]][
		either port/locals/svc [
			control port/locals/svc spec
		][
			install port/locals/scm spec
		]
		port
	]
	
	query: func [port /local ss out][
		either port/locals/svc [
			ss: make struct! SERVICE_STATUS none
			try* [QueryServiceStatus port/locals/svc ss]
			out: make status-class []
			out/state: decode-flags/state ss
			out/type: decode-flags/type ss
			out/exit-code: ss/dwWin32ExitCode
			out/interactive: not zero? ss/dwServiceType and defs/misc/interactive
			out
		][
			list-services port
		]
	]
	
	find: func [port id [string!]][
		check-action 'find		
		if not zero? id: OpenService port/locals/scm id defs/misc/manager-all-access [
			CloseServiceHandle id
		]
		not zero? id
	]
	
	remove: func [port][	
		check-action/svc 'remove
		try* [DeleteService port/locals/svc]
		port
	]
		
	close: func [port /local pl][	
		pl: port/locals
		if pl/svc [
			try* [CloseServiceHandle pl/svc]
		]
		try* [CloseServiceHandle pl/scm]
		port/state/flags: 0
	]
	
	get-modes: func [port spec /local cfg pos direct?][
		check-action/svc 'get-modes
		if none? port/host [
			make error! "no service name specified!"
		]
		if none? port/locals/scm [
			direct?: yes
			port: open port
		]	
		either 'all = first spec [
			change/only next spec copy* next first svc-class
		][
			cfg: get-config port/locals/svc port/host spec
			forskip spec 2 [
				poke spec 2 get in cfg first spec
			]
			if pos: find* spec first [comment:][
				poke pos 2 get-comment port/locals/svc
			]
			new-line/skip head spec true 2
		]
		if direct? [close port]
	]
	
	set-modes: func [port spec /local swc more? direct?][
		check-action/svc 'set-modes
		if none? port/host [
			make error! "no service name specified!"
		]
		if none? port/locals/scm [
			direct?: yes
			port: open port
		]
		swc: first [comment:]
		more?: yes
		if find* spec first [comment:][
			set-comment port/locals/svc spec
			more?: 2 < length? spec
		]
		if more? [set-config port/locals/svc spec]
		if direct? [close port]
	]
	
	copy: func [port][
		either port/locals/svc [
			get-modes* port copy* next first svc-class
		][
			query* port
		]	
	]
		
	net-utils/net-install SCM self none
]

nt-services: context [

	install: func [spec [block!]][
		write scm:// spec
	]
	
	uninstall: func [name [string!]][
		close remove open join scm:// name
	]
	
	start: func [name [string!]][
		write join scm:// name [start]
	]
	
	stop: func [name [string!]][
		write join scm:// name [stop]
	]
	
	exists?: func [name [string!] /local p result][
		result: find p: open scm:// name
		close p
		result
	]
	
	running?: func [name [string!] /local p result][
		result: get in query p: open join scm:// name 'state
		close p
		to logic! find [
			start-pending
			running
			continue-pending
		] result
	]
	
	attach: func [
		name [string!]
		/notify id [integer!] on-quit [block!]
		/local lib ServiceLaunch port
	][
		lib: load/library join first split-path system/options/boot %service.dll
		ServiceLaunch: make routine! [
			name	[string!]
			port-id [integer!]
		] lib "ServiceLaunch"
		
		ServiceLaunch name id
		
		if notify [
			port: open/binary/direct/no-wait join udp://: id
			port/awake: func [p][
				if "Q" = copy p [do on-quit]
				true
			]
			append system/ports/wait-list port
		]
	]
	
	run: :attach ; for backward compatibility
]
